"""Chow classifier package for time-series trend classification."""

# __init__.py

# Version of the chowclassifier package
__version__ = "1.0.9"
from chowclassifier.chow import Chow
from chowclassifier.chewdata import ChewData
